import asyncio
import argparse
import os
from bleak import BleakScanner, BleakClient
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from typing import List

LEDGER_SERVICE_UUID_EUROPA = "13d63400-2c97-3004-0000-4c6564676572"
LEDGER_SERVICE_UUID_STAX = "13d63400-2c97-6004-0000-4c6564676572"
LEDGER_SERVICE_UUID_NANOX = "13d63400-2c97-0004-0000-4c6564676572"

TAG_ID = b"\x05"

def get_argparser():
    parser = argparse.ArgumentParser(description="Manage ledger ble devices.")
    parser.add_argument("--show", help="Show currently selected ledger device.", action='store_true')
    parser.add_argument("--demo", help="Get version demo (connect to ble device, send get version, print response and disconnect).", action='store_true')
    return parser

class NoLedgerDeviceDetected(Exception):
    pass

class BleScanner(object):
    def __init__(self):
        self.devices = []

    def __scan_callback(self, device: BLEDevice, advertisement_data: AdvertisementData):
        if LEDGER_SERVICE_UUID_STAX in advertisement_data.service_uuids or LEDGER_SERVICE_UUID_NANOX in advertisement_data.service_uuids or LEDGER_SERVICE_UUID_EUROPA in advertisement_data.service_uuids:
            device_is_in_list = False
            for dev in self.devices:
                if device.address == dev[0]:
                    device_is_in_list = True
            if not device_is_in_list:
                self.devices.append((device.address, device.name))

    async def scan(self):
        scanner = BleakScanner(
            self.__scan_callback,
        )
        await scanner.start()
        counter = 0
        while counter < 5000:
            await asyncio.sleep(0.01)
            counter += 1
        await scanner.stop()

queue: asyncio.Queue = asyncio.Queue()

def callback(sender, data):
    response = bytes(data)
    queue.put_nowait(response)

async def _get_client(address: str) -> BleakClient:
    # Connect to client
    client = BleakClient(address)
    await client.connect()

    characteristic_notify = None
    characteristic_write_with_rsp = None
    characteristic_write_cmd = None
    for service in client.services:
        if service.uuid in [LEDGER_SERVICE_UUID_NANOX, LEDGER_SERVICE_UUID_STAX, LEDGER_SERVICE_UUID_EUROPA]:
            for char in service.characteristics:
                if "0001" in char.uuid:
                    characteristic_notify = char
                if "0002" in char.uuid:
                    characteristic_write_with_rsp = char
                if "0003" in char.uuid:
                    characteristic_write_cmd = char

    assert characteristic_notify != None

    # Choose write cmd to speed up transfers
    characteristic_write = characteristic_write_cmd
    if characteristic_write == None:
        characteristic_write = characteristic_write_with_rsp

    # Register notifications callback
    await client.start_notify(characteristic_notify, callback)

    # Enable notifications
    await client.write_gatt_char(characteristic_write.uuid, bytes.fromhex("0001"))
    assert await queue.get() == b"\x00\x00\x00\x00\x00"

    # Get MTU value
    await client.write_gatt_char(characteristic_write.uuid, bytes.fromhex("0800000000"))
    response = await queue.get()
    mtu = int.from_bytes(response[5:6], 'big')

    print("[BLE] MTU {:d}".format(mtu))

    return client, mtu, characteristic_write

async def _read(mtu) -> bytes:
    response = await queue.get()
    assert len(response) >= 5
    assert response[0] == TAG_ID[0]
    sequence = int.from_bytes(response[1:3], 'big')
    assert sequence == 0
    total_size = int.from_bytes(response[3:5], "big")

    total_size_bkup = total_size
    if total_size >= (mtu-5):
        apdu = response[5:mtu]
        total_size -= (mtu-5)

        response = await queue.get()
        assert response[0] == TAG_ID[0]
        assert len(response) >= 3
        next_sequence = int.from_bytes(response[1:3], 'big')
        assert next_sequence == sequence+1
        sequence = next_sequence

        while total_size >= (mtu-3):
            apdu += response[3:mtu]
            total_size -= (mtu-3)
            response = await queue.get()
            assert response[0] == TAG_ID[0]
            assert len(response) >= 3
            sequence = int.from_bytes(response[1:3], 'big')
            assert next_sequence == sequence+1
            sequence = next_sequence

        if total_size > 0:
            apdu += response[3:3+total_size]
    else:
        apdu = response[5:]

    assert len(apdu) == total_size_bkup
    return apdu


async def _write(client: BleakClient, data: bytes, write_characteristic, mtu: int = 20):
    chunks: List[bytes] = []
    buffer = data
    while buffer:
        if not chunks:
            size = 5
        else:
            size = 3
        size = mtu - size
        chunks.append(buffer[:size])
        buffer = buffer[size:]

    for i, chunk in enumerate(chunks):
        header = TAG_ID
        header += i.to_bytes(2, "big")
        if i == 0:
            header += len(data).to_bytes(2, "big")
        await client.write_gatt_char(write_characteristic.uuid, header + chunk)


class BleDevice(object):
    def __init__(self, address):
        self.address = address
        self.loop = None
        self.client = None
        self.opened = False
        self.mtu = 20
        self.write_characteristic = None

    def open(self):
        self.loop = asyncio.new_event_loop()
        self.client, self.mtu, self.write_characteristic = self.loop.run_until_complete(_get_client(self.address))
        self.opened = True

    def close(self):
        if self.opened:
            self.loop.run_until_complete(self.client.disconnect())
            self.opened = False
            self.loop.close()

    def __write(self, data: bytes):
        self.loop.run_until_complete(_write(self.client, data, self.write_characteristic, self.mtu))

    def __read(self) -> bytes:
        return self.loop.run_until_complete(_read(self.mtu))

    def exchange(self, data: bytes, timeout=1000) -> bytes:
        self.__write(data)
        return self.__read()

if __name__ == "__main__":
    args = get_argparser().parse_args()
    try:
        if args.show:
            print(f"Environment variable LEDGER_BLE_MAC currently set to '{os.environ['LEDGER_BLE_MAC']}'")
        elif args.demo:
            try:
                ble_device = BleDevice(os.environ['LEDGER_BLE_MAC'])
            except Exception as ex:
                print(f"Please run 'python -m ledgerblue.BleComm' to select wich device to connect to")
                raise ex
            print("-----------------------------Get version BLE demo------------------------------")
            ble_device.open()
            print(f"Connected to {ble_device.address}")
            get_version_apdu = bytes.fromhex("e001000000")
            print(f"[BLE] => {get_version_apdu.hex()}")
            result = ble_device.exchange(get_version_apdu)
            print(f"[BLE] <= {result.hex()}")
            ble_device.close()
            print(f"Disconnected from {ble_device.address}")
            print(79*"-")
        else:
            scanner = BleScanner()
            asyncio.run(scanner.scan())
            devices_str = ""
            device_idx = 0
            if len(scanner.devices):
                for device in scanner.devices:
                    devices_str += f"  -{device_idx+1}- mac=\"{device[0]}\" name=\"{device[1]}\"\n"
                    device_idx += 1
                index = int(input(f"Select device by index\n{devices_str}"))
                print(f"Please run 'export LEDGER_BLE_MAC=\"{scanner.devices[index-1][0]}\"' to select which device to connect to")
            else:
                raise NoLedgerDeviceDetected
    except NoLedgerDeviceDetected as ex:
        print(ex)
    except Exception as ex:
        raise ex




"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

import argparse
import ssl

def get_argparser():
	parser = argparse.ArgumentParser("Update the firmware by using Ledger to open a Secure Channel.")
	parser.add_argument("--url", help="Server URL", default="https://hsmprod.hardwarewallet.com/hsm/process")
	parser.add_argument("--bypass-ssl-check", help="Keep going even if remote certificate verification fails", action='store_true', default=False)
	parser.add_argument("--apdu", help="Display APDU log", action='store_true')
	parser.add_argument("--perso", help="""A reference to the personalization key; this is a reference to the specific
Issuer keypair used by Ledger to sign the device's Issuer Certificate""", default="perso_11")
	parser.add_argument("--firmware", help="A reference to the firmware to load", required=True)
	parser.add_argument("--targetId", help="The device's target ID (default is Ledger Blue)", type=auto_int, default=0x31000002)
	parser.add_argument("--firmwareKey", help="A reference to the firmware key to use", required=True)
	return parser

def auto_int(x):
	return int(x, 0)

def serverQuery(request, url):
	data = request.SerializeToString()
	urll = urlparse.urlparse(args.url)
	req = urllib2.Request(args.url, data, {"Content-type": "application/octet-stream" })
	if args.bypass_ssl_check:
		res = urllib2.urlopen(req, context=ssl._create_unverified_context())
	else:
		res = urllib2.urlopen(req)
	data = res.read()
	response = Response()
	response.ParseFromString(data)
	if len(response.exception) != 0:
		raise Exception(response.exception)
	return response

if __name__ == '__main__':
	import struct
	import urllib.request as urllib2
	import urllib.parse as urlparse
	from .BlueHSMServer_pb2 import Request, Response
	from .comm import getDongle

	args = get_argparser().parse_args()

	dongle = getDongle(args.apdu)

	# Identify

	targetid = bytearray(struct.pack('>I', args.targetId))
	apdu = bytearray([0xe0, 0x04, 0x00, 0x00]) + bytearray([len(targetid)]) + targetid
	dongle.exchange(apdu)

	# Get nonce and ephemeral key

	request = Request()
	request.reference = "distributeFirmware11_scan"
	parameter = request.remote_parameters.add()
	parameter.local = False
	parameter.alias = "persoKey"
	parameter.name = args.perso
	if args.targetId&0xF >= 0x3:
		parameter = request.remote_parameters.add()
		parameter.local = False
		parameter.alias = "scpv2"
		parameter.name = "dummy"
	request.largeStack = True

	response = serverQuery(request, args.url)

	offset = 0

	remotePublicKey = response.response[offset : offset + 65]
	offset += 65
	nonce = response.response[offset : offset + 8]
	if args.targetId&0xF >= 0x3:
		offset += 8
		masterPublicKey = response.response[offset : offset + 65]
		offset += 65
		masterPublicKeySignatureLength = response.response[offset + 1] + 2
		masterPublicKeySignature = response.response[offset : offset + masterPublicKeySignatureLength]

	# Initialize chain

	apdu = bytearray([0xe0, 0x50, 0x00, 0x00, 0x08]) + nonce
	deviceInit = dongle.exchange(apdu)
	deviceNonce = deviceInit[4 : 4 + 8]

	# Get remote certificate

	request = Request()
	request.reference = "distributeFirmware11_scan"
	request.id = response.id
	parameter = request.remote_parameters.add()
	parameter.local = False
	parameter.alias = "persoKey"
	parameter.name = args.perso
	request.parameters = bytes(deviceNonce)
	request.largeStack = True

	response = serverQuery(request, args.url)

	offset = 0

	responseLength = response.response[offset + 1]
	remotePublicKeySignatureLength = responseLength + 2
	remotePublicKeySignature = response.response[offset : offset + remotePublicKeySignatureLength]

	certificate = bytearray([len(remotePublicKey)]) + remotePublicKey + bytearray([len(remotePublicKeySignature)]) + remotePublicKeySignature
	apdu = bytearray([0xE0, 0x51, 0x80, 0x00]) + bytearray([len(certificate)]) + certificate
	dongle.exchange(apdu)

	# Walk the chain

	index = 0
	while True:
			if index == 0:
				certificate = bytearray(dongle.exchange(bytearray.fromhex('E052000000')))
			elif index == 1:
				certificate = bytearray(dongle.exchange(bytearray.fromhex('E052800000')))
			else:
					break
			if len(certificate) == 0:
				break
			request = Request()
			request.reference = "distributeFirmware11_scan"
			request.id = response.id
			request.parameters = bytes(certificate)
			request.largeStack = True
			serverQuery(request, args.url)
			index += 1

	# Commit agreement and send firmware

	request = Request()
	request.reference = "distributeFirmware11_scan"
	if args.targetId&0xF >= 0x3:
		parameter = request.remote_parameters.add()
		parameter.local = False
		parameter.alias = "scpv2"
		parameter.name = "dummy"
	request.id = response.id
	request.largeStack = True

	response = serverQuery(request, args.url)
	responseData = bytearray(response.response)

	dongle.exchange(bytearray.fromhex('E053000000'))

	for i in range(100):
		if len(responseData) == 0:
			break
		if bytes(responseData[0:4]) == b"SECU":
			raise Exception("Security exception " + chr(responseData[4]))

		responseData = dongle.exchange(responseData)

		request = Request()
		request.reference = "distributeFirmware11_scan"
		request.parameters = b"\xFF" + b"\xFF" + bytes(responseData)
		request.id = response.id
		request.largeStack = True

		response = serverQuery(request, args.url)
		responseData = bytearray(response.response)


	request = Request()
	request.reference = "distributeFirmware11_scan"
	parameter = request.remote_parameters.add()
	parameter.local = False
	parameter.alias = "firmware"
	parameter.name = args.firmware
	parameter = request.remote_parameters.add()
	parameter.local = False
	parameter.alias = "firmwareKey"
	parameter.name = args.firmwareKey
	request.id = response.id
	request.largeStack = True

	response = serverQuery(request, args.url)
	responseData = bytearray(response.response)

	offset = 0
	while offset < len(responseData):
		apdu = responseData[offset : offset + 5 + responseData[offset + 4]]
		dongle.exchange(apdu)
		offset += 5 + responseData[offset + 4]

	dongle.close()



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2019 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from .commException import CommException
from binascii import hexlify
import socket
import struct

class DongleServer(object):
	def __init__(self, server, port, debug=False):
		self.server = server
		self.port = port
		self.debug = debug
		self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.opened = True
		try:
			self.socket.connect((self.server, self.port))
		except:
			raise CommException("Proxy connection failed")

	def exchange(self, apdu, timeout=20000):

		def send_apdu(apdu):
			if self.debug:
				print("=> %s" % hexlify(apdu))
			self.socket.send(struct.pack(">I", len(apdu)))
			self.socket.send(apdu)

		def get_data():
			size = struct.unpack(">I", self.socket.recv(4))[0]
			response = self.socket.recv(size)
			sw = struct.unpack(">H", self.socket.recv(2))[0]
			if self.debug:
				print("<= %s%.2x" % (hexlify(response), sw))
			return sw, response


		send_apdu(apdu)
		(sw, response) = get_data()
		if sw == 0x9000:
			return bytearray(response)
		else:
			# handle the get response case:
			# When more data is available, the chip sends 0x61XX
			# So 0x61xx as a SW must not be interpreted as an error
			if (sw & 0xFF00) != 0x6100:
				raise CommException("Invalid status %04x" % sw, sw)
			else:
				while (sw & 0xFF00) == 0x6100:
					send_apdu(bytes.fromhex("00c0000000"))  # GET RESPONSE
					(sw, data) = get_data()
					response += data

				# Check that the last received SW is indeed 0x9000
				if sw == 0x9000:
					return bytearray(response)

		# In any other case return an exception
		raise CommException("Invalid status %04x" % sw, sw)

	def apduMaxDataSize(self):
		return 240

	def close(self):
		try:
			self.socket.close()
			self.socket = None
		except:
			pass
		self.opened = False

def getDongle(server="127.0.0.1", port=9999, debug=False):
    return DongleServer(server, port, debug)



from ledgerblue.comm import getDongle
from ledgerblue.hexLoader import HexLoader
from ledgerblue.recoverMutualAuth import recoverMutualAuth, SCPv3, recoverValidate
from ledgerblue.recoverUtil import Recover
from hashlib import sha256
from Crypto.Cipher import AES
import binascii
import argparse
import json
import gnupg
import os


def get_argparser():
    parser = argparse.ArgumentParser(description="Backup 3 shares of a seed.")
    parser.add_argument("--targetId", help="The device's target ID (default is Nano X)", type=auto_int)
    parser.add_argument("--rootPrivateKey", help="The private key of the Certificate Authority")
    parser.add_argument("--issuerPublicKey", help="The public key of the Issuer (used to verify the certificate of "
                                                  "the device)")
    parser.add_argument("--numberOfWords", help="The number of words in the mnemonic (default is 24 words)", type=auto_int)
    parser.add_argument("-c", "--configuration", help="Configuration file", default=None, required=True, action="store")
    parser.add_argument("--gpg", help="Encrypt the backup data. Enter the email address associated to your gpg key",
                        default=None, action="store")
    return parser


def auto_int(x):
    return int(x, 0)


def decode_bytes(my_bytes):
    return my_bytes.decode('utf-8')


if __name__ == '__main__':

    args = get_argparser().parse_args()

    # Read the configuration file
    try:
        with open(args.configuration,'r') as file:
            conf = json.load(file)
    except Exception as err:
        print(err)
        exit()

    if args.targetId is None:
        args.targetId = 0x33000004
    if args.rootPrivateKey is None:
        raise Exception("Missing Certificate Authority private key")
    if args.issuerPublicKey is None:
        args.issuerPublicKey = '0490f5c9d15a0134bb019d2afd0bf297149738459706e7ac5be4abc350a1f81805' \
                               '7224fce12ec9a65de18ec34d6e8c24db927835ea1692b14c32e9836a75dad609'
    if args.numberOfWords is None:
        # Default is 24 words
        args.numberOfWords = 24
    if not args.gpg:
        print("Your backup is going to be saved unencrypted !")
    else:
        home = os.path.join(os.environ['HOME'], ".gnupg")
        gpg = gnupg.GPG(gnupghome=home)

    dongle = getDongle(True)

    orchestratorPrivateKey = conf['orchestrator']['key']

    # Execute device-orchestrator secure channel protocol
    secret, devicePublicKey = SCPv3(dongle, bytearray.fromhex(args.issuerPublicKey), args.targetId,
                                    bytearray.fromhex(args.rootPrivateKey), bytearray.fromhex(orchestratorPrivateKey))
    loader = HexLoader(dongle, 0xe0, True, secret, scpv3=True)

    # Initialize the session with the info from the configuration file
    recoverSession = Recover(conf)
    confirmed = False

    for provider in conf['providers']:
        backup_data = dict()
        delete_data = dict()
        name = provider['name']
        privateKey = provider['key']

        # Execute device-provider mutual authentication
        providerSk, providerPk = recoverValidate(loader, bytearray.fromhex(args.rootPrivateKey), name,
                                                 bytearray.fromhex(privateKey))
        loader.recoverMutualAuth()
        sharedKey = recoverMutualAuth(providerSk, devicePublicKey)
        recoverSession.sharedKey = sharedKey

        # Send the user identity to the device
        if not confirmed:
            dataIdv = recoverSession.recoverPrepareDataIdv()
            cipher = AES.new(sharedKey, AES.MODE_SIV)
            ciphertext, tag = cipher.encrypt_and_digest(dataIdv)
            loader.recoverConfirmID(tag, ciphertext)
            confirmed = True

        dataHash = recoverSession.recoverPrepareDataHash(providerPk)
        cipher = AES.new(sharedKey, AES.MODE_SIV)
        ciphertext, tag = cipher.encrypt_and_digest(dataHash)

        # Validate backup data hash (device and backup provider agree on the same backup data)
        response = loader.recoverValidateHash(tag, ciphertext)

        # Get the share value and the number of words of the mnemonic (can be 12, 18 or 24)
        # The share is encrypted with the device-provider shared key
        resp = loader.recoverGetShare(value='shares')
        encryptedData = resp[:len(resp)- 1]
        numberOfWords = resp[len(resp) - 1]
        share, idx, deletePublicKey, commitHash = recoverSession.recoverBackupProviderDecrypt(encryptedData)

        # Get the commitments to the coefficients used to calculate the share
        commitments = loader.recoverGetShare(value='commitments')

        # Get the VSS point
        commitmentPoint = loader.recoverGetShare(value='point')

        h = sha256()
        h.update(commitments)
        calculatedCommitHash = h.digest()

        # Hashes of the commitments match on both side
        if not (commitHash == calculatedCommitHash):
            raise Exception("Hashes of the commitments don't match")

        # Verify whether the share is consistent
        result, shareCommit = recoverSession.recoverVerifyCommitments(share, idx, commitments, commitmentPoint)
        if result:
            # Do the backup
            backup_data[name] = dict()
            backup_data[name]['share'] = decode_bytes(binascii.hexlify(share))
            backup_data[name]['index'] = decode_bytes(binascii.hexlify(idx.to_bytes(4, 'little')))
            backup_data[name]['commitments'] = decode_bytes(binascii.hexlify(commitments))
            backup_data[name]['share_commit'] = decode_bytes(binascii.hexlify(shareCommit))
            backup_data[name]['hash'] = decode_bytes(binascii.hexlify(commitHash))
            backup_data[name]['point'] = decode_bytes(binascii.hexlify(commitmentPoint))
            backup_data[name]['words_number'] = numberOfWords
            delete_data[name] = dict()
            delete_data[name]['public_key'] = decode_bytes(binascii.hexlify(deletePublicKey))
        else:
            raise Exception("Share's commitments not verified")

        # Validate the share's commitment 
        recoverSession.recoverValidateShareCommit(loader, shareCommit)

        # Store the backup (it is up to the user to store it in a safe location)
        for name in backup_data:
            try:
                if args.gpg:
                    with open(name + ".json.gpg",'w') as file:
                        encode_backup_data = json.dumps(backup_data, indent = 4, sort_keys = True).encode("utf-8")
                        encrypted_backup = gpg.encrypt(encode_backup_data,recipients=[args.gpg])
                        file.write(str(encrypted_backup))
                    with open("Delete" + name + ".json.gpg",'w') as file:
                        encode_delete_data = json.dumps(delete_data, indent = 4).encode("utf-8")
                        encrypted_data = gpg.encrypt(encode_delete_data,recipients=[args.gpg])
                        file.write(str(encrypted_data))
                else:
                    with open(name + ".json",'w') as file:
                        json.dump(backup_data, file, indent = 4, sort_keys = True)
                    with open("Delete" + name + ".json",'w') as file:
                        json.dump(delete_data, file, indent = 4)
            except Exception as err:
                print(err)
                exit()




"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

class IntelHexArea:
        def __init__(self, start, data):
                self.start = start
                self.data = data

        def getStart(self):
                return self.start

        def getData(self):
                return self.data

        def setData(self, data):
                self.data = data

def insertAreaSorted(areas, area):
        i=0
        while i < len(areas):
                if area.start < areas[i].start:
                        break
                i+=1
        #areas = areas[0:i] + [area]
        areas[i:i] = [area]
        return areas


class IntelHexParser:
        # order by start address
        def _addArea(self, area):
                self.areas = insertAreaSorted(self.areas, area)

        def __init__(self, fileName):
                self.bootAddr = 0
                self.areas = []
                lineNumber = 0
                startZone = None
                startFirst = None
                current = None
                zoneData = b''
                file = open(fileName, "r")
                for data in file:
                        lineNumber += 1
                        data = data.rstrip('\r\n')
                        if len(data) == 0:
                                continue
                        if data[0] != ':':
                                raise Exception("Invalid data at line %d" % lineNumber)
                        data = bytearray.fromhex(data[1:])
                        count = data[0]
                        address = (data[1] << 8) + data[2]
                        recordType = data[3]
                        if recordType == 0x00:
                                if startZone == None:
                                        raise Exception("Data record but no zone defined at line " + str(lineNumber)) 
                                if startFirst == None:
                                        startFirst = address
                                        current = startFirst
                                if address != current:
                                        self._addArea(IntelHexArea((startZone << 16) + startFirst, zoneData))
                                        zoneData = b''
                                        startFirst = address
                                        current = address
                                zoneData += data[4:4 + count]
                                current += count
                        if recordType == 0x01:
                                if len(zoneData) != 0:
                                        self._addArea(IntelHexArea((startZone << 16) + startFirst, zoneData))
                                        zoneData = b''
                                        startZone = None
                                        startFirst = None
                                        current = None
                        if recordType == 0x02:
                                        raise Exception("Unsupported record 02")
                        if recordType == 0x03:
                                        raise Exception("Unsupported record 03")
                        if recordType == 0x04:
                                        if len(zoneData) != 0:
                                                self._addArea(IntelHexArea((startZone << 16) + startFirst, zoneData))
                                                zoneData = b''
                                                startZone = None
                                                startFirst = None
                                                current = None
                                        startZone = (data[4] << 8) + data[5]
                        if recordType == 0x05:
                                        self.bootAddr = ((data[4]&0xFF) << 24) + ((data[5]&0xFF) << 16) + ((data[6]&0xFF) << 8) + (data[7]&0xFF)
                #tail add of the last zone
                if len(zoneData) != 0:
                        self._addArea(IntelHexArea((startZone << 16) + startFirst, zoneData))
                        zoneData = b''
                        startZone = None
                        startFirst = None
                        current = None
                file.close()

        def getAreas(self):
                return self.areas

        def getBootAddr(self):
                return self.bootAddr

        def maxAddr(self):
                addr = 0
                for a in self.areas:
                        if a.start+len(a.data) > addr:
                                addr = a.start+len(a.data)
                return addr

        def minAddr(self):
                addr = 0xFFFFFFFF
                for a in self.areas:
                        if a.start < addr:
                                addr = a.start
                return addr

import binascii

class IntelHexPrinter:
        def addArea(self, startaddress, data, insertFirst=False):
                #self.areas.append(IntelHexArea(startaddress, data))
                if insertFirst:
                        self.areas = [IntelHexArea(startaddress, data)] + self.areas
                else:
                        #order by start address
                        self.areas = insertAreaSorted(self.areas, IntelHexArea(startaddress, data))

        def __init__(self, parser=None, eol="\r\n"):
                self.areas = []
                self.eol = eol
                self.bootAddr = 0
                # build bound to the parser
                if parser:
                        for a in parser.areas:
                                self.addArea(a.start, a.data)
                        self.bootAddr = parser.bootAddr

        def getAreas(self):
                return self.areas

        def getBootAddr(self):
                return self.bootAddr

        def maxAddr(self):
                addr = 0
                for a in self.areas:
                        if a.start+len(a.data) > addr:
                                addr = a.start+len(a.data)
                return addr

        def minAddr(self):
                addr = 0xFFFFFFFF
                for a in self.areas:
                        if a.start < addr:
                                addr = a.start
                return addr

        def setBootAddr(self, bootAddr):
                self.bootAddr = int(bootAddr)

        def checksum(self, bin):
                cks = 0
                for b in bin:
                        cks += b
                cks = (-cks) & 0x0FF
                return cks

        def _emit_binary(self, file, bin):
                cks = self.checksum(bin)
                s = (":" + binascii.hexlify(bin).decode('utf-8') + hex(0x100+cks)[3:] + self.eol).upper()
                if file != None:
                        file.write(s)
                else:
                        print(s)

        def writeTo(self, fileName, blocksize=32):
                file = None
                if fileName != None:
                        file = open(fileName, "w")
                for area in self.areas:
                        off = 0
                        # force the emission of selection record at start
                        oldoff = area.start + 0x10000
                        while off < len(area.data):
                                # emit a offset selection record
                                if (off & 0xFFFF0000) != (oldoff & 0xFFFF0000):
                                        self._emit_binary(file, bytearray.fromhex("02000004" + hex(0x10000+(area.start>>16))[3:7]))

                                # emit data record
                                if off+blocksize > len(area.data):
                                        self._emit_binary(file, bytearray.fromhex(hex(0x100+(len(area.data)-off))[3:] + hex(0x10000+off+(area.start&0xFFFF))[3:] + "00") + area.data[off:len(area.data)])
                                else:
                                        self._emit_binary(file, bytearray.fromhex(hex(0x100+blocksize)[3:] + hex(0x10000+off+(area.start&0xFFFF))[3:] + "00") + area.data[off:off+blocksize])

                                oldoff = off
                                off += blocksize

                bootAddrHex = hex(0x100000000+self.bootAddr)[3:]
                s = ":04000005"+bootAddrHex+hex(0x100+self.checksum( bytearray.fromhex("04000005"+bootAddrHex)))[3:]+self.eol
                if file != None:
                        file.write(s)
                else:
                        print(s)

                s = ":00000001FF"+self.eol

                if file != None:
                        file.write(s)
                else:
                        print(s)

                if file != None:
                        file.close()



from ledgerblue.comm import getDongle
from ledgerblue.hexLoader import HexLoader
from ledgerblue.recoverMutualAuth import recoverMutualAuth, SCPv3, recoverValidate
from ledgerblue.recoverUtil import Recover
from Crypto.Cipher import AES
import argparse
import os
import json
import gnupg


def get_argparser():
    parser = argparse.ArgumentParser(description="Delete shares backups")
    parser.add_argument("--targetId", help="The device's target ID (default is Nano X)", type=auto_int)
    parser.add_argument("--rootPrivateKey", help="The private key of the Certificate Authority")
    parser.add_argument("--issuerPublicKey", help="The public key of the Issuer (used to verify the certificate of "
                                                  "the device)")
    parser.add_argument("-s", "--select", help="Select the backups to use", required = True, action="store",
                        dest="backups", type=str, nargs='*')
    parser.add_argument("-c", "--configuration", help="Configuration file", default=None, required=True, action="store")
    parser.add_argument("--gpg", help="Decrypt the deletion public key. Enter the email address "
                                      "associated to your gpg key", default=None, action="store")
    return parser


def auto_int(x):
    return int(x, 0)


if __name__ == '__main__':

    args = get_argparser().parse_args()

    # Read the configuration file
    try:
        with open(args.configuration,'r') as file:
            conf = json.load(file)
    except Exception as err:
        print(err)
        exit()

    if args.targetId is None:
        args.targetId = 0x33000004
    if args.rootPrivateKey is None:
        raise Exception("Missing Certificate Authority private key")
    if args.issuerPublicKey is None:
        args.issuerPublicKey = '0490f5c9d15a0134bb019d2afd0bf297149738459706e7ac5be4abc350a1f81805' \
                               '7224fce12ec9a65de18ec34d6e8c24db927835ea1692b14c32e9836a75dad609'
    if not args.gpg:
        print("Reading your deletion public key from unencrypted files")
    else:
        home = os.path.join(os.environ['HOME'], ".gnupg")
        gpg = gnupg.GPG(gnupghome=home)

    dongle = getDongle(True)

    orchestratorPrivateKey = conf['orchestrator']['key']

    # Execute device-orchestrator secure channel protocol
    secret, devicePublicKey = SCPv3(dongle, bytearray.fromhex(args.issuerPublicKey), args.targetId,
                                    bytearray.fromhex(args.rootPrivateKey),
                                    bytearray.fromhex(orchestratorPrivateKey))
    loader = HexLoader(dongle, 0xe0, True, secret, scpv3=True)

    # Initialize the session with the info from the configuration file
    recoverSession = Recover(conf)
    confirmed = False

    for backup in args.backups:
        # Read the deletion public key
        try:
            with open(backup, 'r') as file:
                if args.gpg:
                    enc_delete_data = file.read()
                    dec_data = gpg.decrypt(enc_delete_data)
                    delete_data = json.loads(dec_data.data)
                else:
                    delete_data = json.load(file)
        except Exception as err:
            print(err)
            exit()

        for p in conf['providers']:
            if p['name'] == list(delete_data.keys())[0]:
                break
        name = p['name']
        privateKey = p['key']
        backupPublicKey = delete_data[name]['public_key']

        # Execute device-provider mutual authentication
        providerSk, providerPk = recoverValidate(loader, bytearray.fromhex(args.rootPrivateKey), name,
                                                 bytearray.fromhex(privateKey))
        loader.recoverMutualAuth()
        sharedKey = recoverMutualAuth(providerSk, devicePublicKey)
        recoverSession.sharedKey = sharedKey

        # Send the user identity to the device
        if not confirmed:
            dataIdv = recoverSession.recoverPrepareDataIdv(delete=True)
            cipher = AES.new(sharedKey, AES.MODE_SIV)
            ciphertext, tag = cipher.encrypt_and_digest(dataIdv)
            loader.recoverConfirmID(tag, ciphertext)
            confirmed = True

        dataHash = recoverSession.recoverPrepareDataHash(providerPk)
        cipher = AES.new(sharedKey, AES.MODE_SIV)
        ciphertext, tag = cipher.encrypt_and_digest(dataHash)

        # Validate backup data hash (device and backup provider agree on the same backup data)
        response = loader.recoverValidateHash(tag, ciphertext)

        # Verify whether the backup can be deleted
        recoverSession.recoverDeleteBackup(loader, bytearray.fromhex(backupPublicKey))



from ledgerblue.ecWrapper import PublicKey, PrivateKey
from ledgerblue.hexLoader import HexLoader
from ledgerblue.recoverSCP import scp_derive_key, extract_from_certificate, decrypt_certificate
import binascii
import struct
import os

CERT_ROLE_SIGNER = 0x1
CERT_ROLE_SIGNER_EPHEMERAL = 0x11
CERT_ROLE_DEVICE = 0x2
CERT_ROLE_DEVICE_EPHEMERAL = 0x12
CERT_ROLE_RECOVER_ORCHESTRATOR = 0x05
CERT_ROLE_RECOVER_ORCHESTRATOR_EPHEMERAL = 0x15

CERT_ROLE_RECOVER_PROVIDER = 0x4
CERT_ROLE_RECOVER_PROVIDER_EPHEMERAL = 0x14
CERT_FORMAT_VERSION = 0x01


def SCPv3(dongle, issuer_public_key, target_id, ca_private_key=None, signer_private_key=None,
          user_mode=False):
    if not user_mode:
        ca_sk = PrivateKey(bytes(ca_private_key))
    target = bytearray(struct.pack('>I', target_id))
    scpv3 = 0x02

    apdu = bytearray([0xe0, 0x04, scpv3, 0x00]) + bytearray([len(target)]) + target
    dongle.exchange(apdu)

    # Initialize authentication
    nonce = os.urandom(8)
    apdu = bytearray([0xe0, 0x50, 0x00, 0x00]) + bytearray([len(nonce)]) + nonce

    auth_info = dongle.exchange(apdu)
    device_nonce = auth_info[4:12]

    if user_mode:
        role = CERT_ROLE_SIGNER
        ephemeral_role = CERT_ROLE_SIGNER_EPHEMERAL
    else:
        role = CERT_ROLE_RECOVER_ORCHESTRATOR
        ephemeral_role = CERT_ROLE_RECOVER_ORCHESTRATOR_EPHEMERAL

    # Validate signer static certificate
    certificate_id = 0x00
    if signer_private_key is not None:
        signer_sk = PrivateKey(bytes(signer_private_key))
        signer_pk = bytearray(signer_sk.pubkey.serialize(compressed=False))
        data_to_sign = bytes(bytearray([role]) + signer_pk)
        if user_mode:
            signature = signer_sk.ecdsa_sign(bytes(data_to_sign))
            signature = signer_sk.ecdsa_serialize(signature)
        else:
            signature = ca_sk.ecdsa_sign(bytes(data_to_sign))
            signature = ca_sk.ecdsa_serialize(signature)
            certificate_id = 0x01
        certificate = bytearray([len(signer_pk)]) + signer_pk + bytearray([len(signature)]) + signature
        apdu = bytearray([0xE0, 0x51, 0x00, certificate_id]) + bytearray([len(certificate)]) + certificate

        dongle.exchange(apdu)

    # Validate signer ephemeral certificate
    ephemeral_private = PrivateKey()
    ephemeral_public = bytearray(ephemeral_private.pubkey.serialize(compressed=False))
    data_to_sign = bytes(bytearray([ephemeral_role]) + nonce + device_nonce + ephemeral_public)
    signature = signer_sk.ecdsa_sign(bytes(data_to_sign))
    signature = signer_sk.ecdsa_serialize(signature)
    certificate = bytearray([len(ephemeral_public)]) + ephemeral_public + bytearray([len(signature)]) + signature
    apdu = bytearray([0xE0, 0x51, 0x80, certificate_id]) + bytearray([len(certificate)]) + certificate

    dongle.exchange(apdu)

    # Get device certificates
    issuer_pk = PublicKey(bytes(issuer_public_key), raw=True)

    encrypted_certificate_static = dongle.exchange(bytearray.fromhex('E052000000'))

    # First extract the device ephemeral public key from the device ephemeral certificate
    certificate_ephemeral = bytearray(dongle.exchange(bytearray.fromhex('E052800000')))

    certificate_header, certificate_public_key, certificate_signature_array = \
        extract_from_certificate(certificate_ephemeral)

    # Check the certificate's header
    if not certificate_header == bytearray():
        raise Exception("Device ephemeral certificate: error format")

    # Decrypt the device static certificate
    # Compute the shared key
    pub_key = PublicKey(bytes(certificate_public_key), raw=True)
    secret = pub_key.ecdh(binascii.unhexlify(ephemeral_private.serialize()), True)
    key = scp_derive_key(secret, 0, True)

    certificate_static = decrypt_certificate(encrypted_certificate_static, key)

    # Decrypt the signature from the ephemeral certificate
    certificate_signature_array = decrypt_certificate(certificate_signature_array, key)

    # Verify the device static certificate
    certificate_static_header, certificate_static_public_key, certificate_static_signature_array = \
        extract_from_certificate(certificate_static)
    certificate_signature = issuer_pk.ecdsa_deserialize(bytes(certificate_static_signature_array))
    certificate_signed_data = bytearray([CERT_ROLE_DEVICE]) + certificate_static_header + certificate_static_public_key

    if not issuer_pk.ecdsa_verify(bytes(certificate_signed_data), certificate_signature):
        raise Exception("Device certificate not verified")

    # Verify the device ephemeral certificate
    device_pub_key = PublicKey(bytes(certificate_static_public_key), raw=True)
    certificate_signature = device_pub_key.ecdsa_deserialize(bytes(certificate_signature_array))
    certificate_signed_data = bytearray([CERT_ROLE_DEVICE_EPHEMERAL]) + device_nonce + nonce + certificate_public_key

    if not device_pub_key.ecdsa_verify(bytes(certificate_signed_data), certificate_signature):
        raise Exception("Device ephemeral certificate not verified")

    dongle.exchange(bytearray.fromhex('E053000000'))

    return secret, certificate_public_key


def recoverValidate(loader, caKey, name, staticPrivateKey):
    caPrivateKey = PrivateKey(bytes(caKey))

    providerPrivateKey = PrivateKey(bytes(staticPrivateKey))
    providerPublicKey = bytearray(providerPrivateKey.pubkey.serialize(compressed=False))

    # Validate provider's static certificate
    role = bytearray([CERT_ROLE_RECOVER_PROVIDER])
    version = bytearray([CERT_FORMAT_VERSION])
    dataToSign = bytes(version + role + struct.pack('>B', len(name)) + name.encode() +
                       struct.pack('>B', len(providerPublicKey)) + providerPublicKey)
    signature = caPrivateKey.ecdsa_sign(bytes(dataToSign))
    signature = caPrivateKey.ecdsa_serialize(signature)

    loader.recoverValidateCertificate(bytes(version), bytes(role), name, providerPublicKey, signature)

    # Validate provider's ephemeral certificate
    ephemeralPrivate = PrivateKey()
    ephemeralPublic = bytearray(ephemeralPrivate.pubkey.serialize(compressed=False))
    role = bytearray([CERT_ROLE_RECOVER_PROVIDER_EPHEMERAL])
    dataToSign = bytes(version + role + struct.pack('>B', len(name)) + name.encode() +
                       struct.pack('>B', len(ephemeralPublic)) + ephemeralPublic)
    signature = providerPrivateKey.ecdsa_sign(bytes(dataToSign))
    signature = providerPrivateKey.ecdsa_serialize(signature)

    loader.recoverValidateCertificate(bytes(version), bytes(role), name, ephemeralPublic, signature, True)

    return ephemeralPrivate, ephemeralPublic


def recoverMutualAuth(privateKey, devicePublicKey):

    # Compute the shared key
    pk = PublicKey(devicePublicKey, raw=True)
    secret = pk.ecdh(binascii.unhexlify(privateKey.serialize()), True)
    sharedKey = scp_derive_key(secret, 0, True)

    return sharedKey




"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

import argparse

def get_argparser():
	parser = argparse.ArgumentParser(description="""
.. warning::

   Using this script undermines the security of the device. Caveat emptor.
""")
	parser.add_argument("--apdu", help="Display APDU log", action='store_true')
	parser.add_argument("--id", help="Identity to initialize", type=auto_int, choices=(0, 1, 2), required=True)
	parser.add_argument("--pin", help="Set a PINs to backup the seed for future use")
	parser.add_argument("--prefix", help="Derivation prefix")
	parser.add_argument("--passphrase", help="Derivation passphrase")
	parser.add_argument("--words", help="Derivation phrase")
	return parser

def auto_int(x):
	return int(x, 0)

if __name__ == '__main__':
	import getpass
	import unicodedata

	from .comm import getDongle

	args = get_argparser().parse_args()

	dongle = getDongle(args.apdu)

	def enter_if_none_and_normalize(hint, strg):
		if strg is None: # or len(string) == 0: len 0 is accepted, to specify without being bothered by a message
			strg = getpass.getpass(hint)
		if len(strg) != 0 :
			strg = unicodedata.normalize('NFKD', u''+strg)
		return strg

	if args.id < 2:
		args.pin = enter_if_none_and_normalize("PIN: ", args.pin)
		if args.pin is None or len(args.pin) == 0:
			raise Exception("Missing PIN for persistent identity")
	elif not args.pin is None:
		raise Exception("Can't set a PIN for the temporary identity")

	args.prefix = enter_if_none_and_normalize("Derivation prefix: ", args.prefix)
	args.passphrase = enter_if_none_and_normalize("Derivation passphrase: ", args.passphrase)
	args.words = enter_if_none_and_normalize("Derivation phrase: ", args.words)

	if args.pin:
		apdudata = bytearray([len(args.pin)]) + bytearray(args.pin, 'utf8')
	else:
		apdudata = bytearray([0])

	if args.prefix:
		apdudata += bytearray([len(args.prefix)]) + bytearray(args.prefix, 'utf8')
	else:
		apdudata += bytearray([0])

	if args.passphrase:
		apdudata += bytearray([len(args.passphrase)]) + bytearray(args.passphrase, 'utf8')
	else:
		apdudata += bytearray([0])

	if args.words:
		apdudata += bytearray([len(args.words)]) + bytearray(args.words, 'utf8')
	else:
		apdudata += bytearray([0])

	apdu = bytearray([0xE0, 0xD0, args.id, 0x00, len(apdudata)]) + apdudata
	dongle.exchange(apdu, timeout=3000)

	dongle.close()


"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

DEFAULT_ALIGNMENT = 1024
PAGE_ALIGNMENT = 64

import argparse
import os

NOCRC=False
if "NOCRC" in os.environ and len(os.environ["NOCRC"]) != 0:
	NOCRC=os.environ["NOCRC"]


def get_argparser():
	parser = argparse.ArgumentParser(description="Load an app onto the device from a hex file.")
	parser.add_argument("--targetId", help="The device's target ID (default is Ledger Blue)", type=auto_int, default=0x31000002)
	parser.add_argument("--targetVersion", help="Set the chip target version")
	parser.add_argument("--apiLevel", help="Set the API level of the SDK used to build the app", type=auto_int, default=-1)
	parser.add_argument("--fileName", help="The application hex file to be loaded onto the device", required=True)
	parser.add_argument("--icon", help="The icon content to use (hex encoded)")
	parser.add_argument("--curve", help="""A curve on which BIP 32 derivation is locked ("secp256k1", "secp256r1",
"ed25519" or "bls12381g1"), can be repeated""", action='append')
	parser.add_argument("--path", help="""A BIP 32 path to which derivation is locked (format decimal a'/b'/c), can be
repeated""", action='append')
	parser.add_argument("--path_slip21", help="""A SLIP 21 path to which derivation is locked""", action='append')
	parser.add_argument("--appName", help="The name to give the application after loading it", required=True)
	parser.add_argument("--signature", help="A signature of the application (hex encoded)")
	parser.add_argument("--signApp", help="Sign application with provided signPrivateKey", action='store_true')
	parser.add_argument("--appFlags", help="The application flags", type=auto_int, default=0)
	parser.add_argument("--bootAddr", help="The application's boot address", type=auto_int)
	parser.add_argument("--rootPrivateKey", help="""The Signer private key used to establish a Secure Channel (otherwise
a random one will be generated)""")
	parser.add_argument("--signPrivateKey", help="Set the private key used to sign the loaded app")
	parser.add_argument("--apdu", help="Display APDU log", action='store_true')
	parser.add_argument("--deployLegacy", help="Use legacy deployment API", action='store_true')
	parser.add_argument("--delete", help="Delete the app with the same name before loading the provided one", action='store_true')
	parser.add_argument("--params", help="Store icon and install parameters in a parameter section before the code", action='store_true')
	parser.add_argument("--tlv", help="Use install parameters for all variable length parameters", action='store_true')
	parser.add_argument("--dataSize", help="The code section's size in the provided hex file (to separate data from code, if not provided the whole allocated NVRAM section for the application will remain readonly.", type=auto_int)
	parser.add_argument("--appVersion", help="The application version (as a string)")
	parser.add_argument("--offline", help="Request to only output application load APDUs into given filename")
	parser.add_argument("--offlineText", help="Request to only output application load APDUs into given filename in text mode", action='store_true')
	parser.add_argument("--installparamsSize", help="The loaded install parameters section size (when parameters are already included within the .hex file.", type=auto_int)
	parser.add_argument("--tlvraw", help="Add a custom install param with the hextag:hexvalue encoding", action='append')
	parser.add_argument("--dep", help="Add a dependency over an appname[:appversion]", action='append')
	parser.add_argument("--nocrc", help="Skip CRC generation when loading", action='store_true')

	return parser

def auto_int(x):
	return int(x, 0)

def parse_bip32_path(path):
	import struct
	if len(path) == 0:
		return b""
	elements = path.split('/')
	result = struct.pack('>B', len(elements))
	for pathElement in elements:
		element = pathElement.split('\'')
		if len(element) == 1:
			result = result + struct.pack(">I", int(element[0]))
		else:
			result = result + struct.pack(">I", 0x80000000 | int(element[0]))
	return result

def parse_slip21_path(path):
	import struct
	result = struct.pack('>B', 0x80 | (len(path) + 1))
	result = result + b'\x00' + string_to_bytes(path)
	return result

def string_to_bytes(x):
	return bytes(x, 'ascii')


if __name__ == '__main__':
	from .ecWrapper import PrivateKey
	from .comm import getDongle
	from .hexParser import IntelHexParser, IntelHexPrinter
	from .hexLoader import HexLoader
	from .hexLoader import *
	from .deployed import getDeployedSecretV1, getDeployedSecretV2
	import struct
	import binascii

	args = get_argparser().parse_args()

	if args.rootPrivateKey == None:
		privateKey = PrivateKey()
		publicKey = binascii.hexlify(privateKey.pubkey.serialize(compressed=False))
		print("Generated random root public key : %s" % publicKey)
		args.rootPrivateKey = privateKey.serialize()

	args.appName = string_to_bytes(args.appName)

	parser = IntelHexParser(args.fileName)
	if args.bootAddr == None:
		args.bootAddr = parser.getBootAddr()

	path = b""
	curveMask = 0xff
	if args.curve != None:
		curveMask = 0x00
		for curve in args.curve:
			if curve == 'secp256k1':
				curveMask |= 0x01
			elif curve == 'secp256r1':
				curveMask |= 0x02
			elif curve == 'ed25519':
				curveMask |= 0x04
			elif curve == 'bls12381g1':
				curveMask |= 0x10
			else:
				raise Exception("Unknown curve " + curve)

	if args.path_slip21 != None:
		curveMask |= 0x08
	path += struct.pack('>B',curveMask)
	if args.path != None:
		for item in args.path:
			if len(item) != 0:
				path += parse_bip32_path(item)
	if args.path_slip21 != None:
		for item in args.path_slip21:
			if len(item) != 0:
				path += parse_slip21_path(item)
		if (args.path == None) or ((len(args.path) == 1) and (len(args.path[0]) == 0)):
			path += struct.pack('>B', 0) # Unrestricted, authorize all paths for regular derivation

	if not args.icon is None:
		args.icon = bytearray.fromhex(args.icon)

	signature = None
	if not args.signature is None:
		signature = bytearray.fromhex(args.signature)

	#prepend app's data with the icon content (could also add other various install parameters)
	printer = IntelHexPrinter(parser)

	# Use of Nested Encryption Key within the SCP protocol is mandartory for upgrades
	cleardata_block_len=None
	if args.appFlags & 2:
		# Not true for scp < 3
		# if signature is None:
		# 	raise BaseException('Upgrades must be signed')

		# ensure data can be decoded with code decryption key without troubles.
		cleardata_block_len = 16

	dongle = None
	secret = None
	if not args.offline:
		dongle = getDongle(args.apdu)
		if args.deployLegacy:
			secret = getDeployedSecretV1(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
		else:
			secret = getDeployedSecretV2(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId)
	else:
		fileTarget = open(args.offline, "wb")
		class FileCard:
			def __init__(self, target):
				self.target = target
			def exchange(self, apdu):
				if args.apdu:
					print(binascii.hexlify(apdu))
				apdu = binascii.hexlify(apdu)
				self.target.write(apdu + '\n'.encode())
				return bytearray([])
			def apduMaxDataSize(self):
				# ensure to allow for encryption of those apdu afterward
				return 240
			def close(self):
				self.target.close()
		dongle = FileCard(fileTarget)

	loader = HexLoader(dongle, 0xe0, not args.offline, secret, cleardata_block_len=cleardata_block_len)

	#tlv mode does not support explicit by name removal, would require a list app before to identify the hash to be removed
	if (not (args.appFlags & 2)) and args.delete:
			loader.deleteApp(args.appName)

	if args.tlv:
		#if code length is not provided, then consider the whole provided hex file is the code and no data section is split
		code_length = printer.maxAddr() - printer.minAddr()
		if not args.dataSize is None:
			code_length -= args.dataSize
		else:
			args.dataSize = 0

		installparams = b""

		# express dependency
		if args.dep:
			for dep in args.dep:
				appname = dep
				appversion = None
				# split if version is specified
				if dep.find(":") != -1:
					(appname,appversion) = dep.split(":")
				depvalue = encodelv(string_to_bytes(appname))
				if appversion:
					depvalue += encodelv(string_to_bytes(appversion))
				installparams += encodetlv(BOLOS_TAG_DEPENDENCY, depvalue)

		#add raw install parameters as requested
		if args.tlvraw:
			for tlvraw in args.tlvraw:
				(hextag,hexvalue) = tlvraw.split(":")
				installparams += encodetlv(int(hextag, 16), binascii.unhexlify(hexvalue))

		if (not (args.appFlags & 2)) and ( args.installparamsSize is None or args.installparamsSize == 0 ):
			#build install parameters
			#mandatory app name
			installparams += encodetlv(BOLOS_TAG_APPNAME, args.appName)
			if not args.appVersion is None:
				installparams += encodetlv(BOLOS_TAG_APPVERSION, string_to_bytes(args.appVersion))
			if not args.icon is None:
				installparams += encodetlv(BOLOS_TAG_ICON, bytes(args.icon))
			if len(path) > 0:
				installparams += encodetlv(BOLOS_TAG_DERIVEPATH, path)

			# append install parameters to the loaded file
			param_start = printer.maxAddr()+(PAGE_ALIGNMENT-(args.dataSize%PAGE_ALIGNMENT))%PAGE_ALIGNMENT
			# only append install param section when not an upgrade as it has already been computed in the encrypted and signed chunk
			printer.addArea(param_start, installparams)
			paramsSize = len(installparams)
		else:
			paramsSize = args.installparamsSize
			# split code and install params in the code
			code_length -= args.installparamsSize
		# create app
		#ensure the boot address is an offset
		if args.bootAddr > printer.minAddr():
			args.bootAddr -= printer.minAddr()
		loader.createApp(code_length, args.apiLevel, args.dataSize, paramsSize, args.appFlags, args.bootAddr|1)
	elif args.params:
		paramsSectionContent = []
		if not args.icon is None:
			paramsSectionContent = args.icon
		#take care of aligning the parameters sections to avoid possible invalid dereference of aligned words in the program nvram.
		#also use the default MPU alignment
		param_start = printer.minAddr()-len(paramsSectionContent)-(DEFAULT_ALIGNMENT-(len(paramsSectionContent)%DEFAULT_ALIGNMENT))
		printer.addArea(param_start, paramsSectionContent)
		# account for added regions (install parameters, icon ...)
		appLength = printer.maxAddr() - printer.minAddr()
		loader.createAppNoInstallParams(args.appFlags, appLength, args.appName, None, path, 0, len(paramsSectionContent), string_to_bytes(args.appVersion))
	else:
		# account for added regions (install parameters, icon ...)
		appLength = printer.maxAddr() - printer.minAddr()
		loader.createAppNoInstallParams(args.appFlags, appLength, args.appName, args.icon, path, None, None, string_to_bytes(args.appVersion))


	hash = loader.load(0x0, 0xF0, printer, targetId=args.targetId, targetVersion=args.targetVersion, doCRC=not (args.nocrc or NOCRC))

	print("Application full hash : " + hash)

	if signature == None and args.signApp:
		masterPrivate = PrivateKey(bytes(bytearray.fromhex(args.signPrivateKey)))
		signature = masterPrivate.ecdsa_serialize(masterPrivate.ecdsa_sign(bytes(binascii.unhexlify(hash)), raw=True))
		print("Application signature: " + str(binascii.hexlify(signature)))

	if args.tlv:
		loader.commit(signature)
	else:
		loader.run(args.bootAddr-printer.minAddr(), signature)

	dongle.close()



from ecpy.curves import Curve, Point
from Crypto.Cipher import AES
from ledgerblue.ecWrapper import PublicKey
from ledgerblue.vss import PedersenVSS
from hashlib import sha256
import os
import struct

FIRST_NAME_TAG = 0x20
NAME_TAG = 0x21
DATE_OF_BIRTH_TAG = 0x22
PLACE_OF_BIRTH_TAG = 0x23
BACKUP_FLOW = 0x01
RESTORE_FLOW = 0x10
DELETE_FLOW = 0x11


class Recover:
    def __init__(self, conf):
        self.sharedKey = bytes()
        self.backupId = bytearray.fromhex(conf['backup_info']['backup_id'])
        self.backupName = conf['backup_info']['backup_name']
        user = conf['user_info']
        self.firstName = user['first_name']
        self.lastName = user['last_name']
        self.birthDate = user['birth']
        self.birthPlace = user['city']
        self.userInfo = self.firstName + self.lastName + self.birthDate + self.birthPlace
        self.f_tag = FIRST_NAME_TAG
        self.n_tag = NAME_TAG
        self.d_tag = DATE_OF_BIRTH_TAG
        self.c_tag = PLACE_OF_BIRTH_TAG
        self.VSS = PedersenVSS(Curve.get_curve("secp384r1"))

    def recoverRestoreSeed(self, loader, share, wordsNumber):
        cipher = AES.new(self.sharedKey, AES.MODE_SIV)
        ciphertext, tag = cipher.encrypt_and_digest(share)
        loader.recoverRestoreSeed(tag, ciphertext, wordsNumber)

    def recoverValidateCommit(self, loader, commits, shareCommit):
        dataToHash = bytes(commits + shareCommit)
        h = sha256()
        h.update(dataToHash)
        dataHash = h.digest()
        cipher = AES.new(self.sharedKey, AES.MODE_SIV)
        ciphertext, tag = cipher.encrypt_and_digest(dataHash)
        loader.recoverValidateCommit(0x2, commits)
        loader.recoverValidateCommit(0x4, shareCommit)
        loader.recoverValidateCommit(0x3, None, tag, ciphertext)

    def recoverShareCommit(self, point, share):
        Q = Point(int.from_bytes(point[:self.VSS.domain_len], 'big'),
                  int.from_bytes(point[self.VSS.domain_len:2 * self.VSS.domain_len], 'big'), self.VSS.curve)
        P = self.VSS.pedersen_share_commit(Q, share)
        return bytearray(P.x.to_bytes(self.VSS.domain_len, 'big') + P.y.to_bytes(self.VSS.domain_len, 'big'))

    def recoverValidateShareCommit(self, loader, shareCommit):
        dataToHash = bytes(shareCommit)
        h = sha256()
        h.update(dataToHash)
        dataHash = h.digest()
        cipher = AES.new(self.sharedKey, AES.MODE_SIV)
        ciphertext, tag = cipher.encrypt_and_digest(dataHash)
        loader.recoverValidateCommit(0x4, shareCommit)
        loader.recoverValidateCommit(0x3, None, tag, ciphertext)

    def recoverVerifyCommitments(self, share, idx, commitments, point):
        point_len = 2 * self.VSS.domain_len
        commitsPoints = [Point(int.from_bytes(commitments[i * point_len:i * point_len + self.VSS.domain_len], 'big'),
                                int.from_bytes(commitments[i * point_len + self.VSS.domain_len: i * point_len + 2 * self.VSS.domain_len],
                                               'big'), self.VSS.curve) for i in range(2)]
        Q = Point(int.from_bytes(point[:self.VSS.domain_len], 'big'),
                  int.from_bytes(point[self.VSS.domain_len:2 * self.VSS.domain_len], 'big'), self.VSS.curve)
        result, shareCommitPoint = self.VSS.pedersen_verify_commit(Q, share, idx, commitsPoints)
        shareCommit = bytearray(shareCommitPoint.x.to_bytes(self.VSS.domain_len, 'big') + shareCommitPoint.y.to_bytes(self.VSS.domain_len, 'big'))
        return result, shareCommit

    def recoverPrepareDataHash(self, publicKey):
        backupDataToHash = bytes(self.backupName.encode() + self.userInfo.encode())
        h1 = sha256()
        h1.update(backupDataToHash)
        backupDataHash = h1.digest()
        dataToHash = bytes(publicKey + self.backupId + backupDataHash)
        h2 = sha256()
        h2.update(dataToHash)
        dataHash = h2.digest()

        return dataHash

    def recoverBackupProviderDecrypt(self, encryptedData):
        cipher = AES.new(self.sharedKey, AES.MODE_SIV)
        plaintext = cipher.decrypt_and_verify(encryptedData[16:], encryptedData[:16])

        share = plaintext[:96]
        idx = int.from_bytes(plaintext[96:100], 'little')
        deletePublicKey = plaintext[100:165]
        commitHash = plaintext[165:]

        return share, idx, deletePublicKey, commitHash

    def recoverDeleteBackup(self, loader, backupPublicKey):
        nonce = os.urandom(16)
        cipher = AES.new(self.sharedKey, AES.MODE_SIV)
        ciphertext, tag = cipher.encrypt_and_digest(nonce)
        encryptedSignature = loader.recoverDeleteBackup(tag, ciphertext)
        cipher = AES.new(self.sharedKey, AES.MODE_SIV)
        signature = cipher.decrypt_and_verify(encryptedSignature[16:], encryptedSignature[:16])
        verifyKey = PublicKey(bytes(backupPublicKey), raw=True)
        signature = verifyKey.ecdsa_deserialize(signature)
        if not verifyKey.ecdsa_verify(nonce, signature):
            raise Exception("Invalid signature")

    def recoverPrepareDataIdv(self, delete=False):
        identifiers = [
            (self.backupName, None),
            (self.firstName, self.f_tag),
            (self.lastName, self.n_tag),
            (self.birthDate, self.d_tag),
            (self.birthPlace, self.c_tag),
        ]
        flow = b'\x01' if delete else b'\x00'

        def pack(tup):
            identifier, tag = tup
            data = struct.pack(">B", tag) if tag is not None else b""
            return data + struct.pack(">B", len(identifier.encode())) + identifier.encode()

        return self.backupId + b"".join(map(pack, identifiers)) + flow

